/*-*-linux-c-*-*/

/*
 *  gnewtellium - Newtella for Unix
 *  Copyright (C) 2001 Elias Athanasopoulos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>

#include "global.h"
#include "gui.h"
#include "newtella.h"
#include "connection.h"

#ifdef SOLARIS
	#include <sys/filio.h>
#endif

extern struct newtella_options *gl_options;
extern int gl_server;
extern struct host_info *gl_my_host_info;

int bytes_avail(int s)
{
	int ret;

	ioctl(s, FIONREAD, &ret);

	return ret;
}

int send_packet(int s, const char *buf, int len)
{
	int rval = send(s, buf, len, 0);

	if (rval == -1) 
	{
		perror("send");
#ifdef SOLARIS
		/* FIXME: ENOTCONN is ok only if con->state == GS_WAITINGCONNECT */
		if (!(errno == ENOTCONN || errno == EAGAIN))
#else
		if (errno != EAGAIN) 
#endif
			return(-1);
		
	}

	return(rval);
}

char *recv_packet(int s, int len)
{
	char *buf = newtella_malloc(len);

	int rval = recv(s, buf, len, 0);

	if (rval == -1) 
	{
		perror("recv_packet");
		return(NULL);
	}

	return(buf);
}

int con_recv_packet(struct newtella_connection *con, int len)
{
  	int rval = recv(con->s, con->b, len, 0);
	
	if (rval == 0 || (rval < 0 && errno != EAGAIN)) {
		perror("con_recv");
		/* con->state = GS_FAILED;*/
		return 0;
	} else 
		return (rval);

}

int con_send_packet(struct newtella_connection *con, const char *buf, int len)
{
	int rval;
	packet *packet_to_send; 

	rval = send(con->s, buf, len, 0);

	if (rval < 0) 
	{ 
#ifdef SOLARIS
		if (!((con->state == GS_WAITINGCONNECT && errno == ENOTCONN) || errno == EAGAIN)) {
#else
		if (errno != EAGAIN) {
#endif
			con->total_dropped_to_send++;
			con->state = GS_FAILED;
			gui_update_con_status(con);
			perror("con_send_packet");
			/* con_remove(con); */
			return(-1);
		} 
		packet_to_send = newtella_malloc(sizeof(packet));
		packet_to_send->_time = time((time_t *) NULL);
		packet_to_send->len = len;
		memcpy(packet_to_send->buf, buf, len);
		con->packet_queue = g_list_append(con->packet_queue, packet_to_send);
		return (0);
	}
       
	con->total_sent++;
	return (rval);
}

int con_send_from_queue(struct newtella_connection *con)
{
	GList *iter;
	packet *packet_to_send;

	if (g_list_length(con->packet_queue) == 0)
		return -1;
	
	iter = g_list_first(con->packet_queue);

	while (iter != NULL) {
		packet_to_send = iter->data;
		if (send(con->s, packet_to_send->buf, packet_to_send->len, 0) != -1) {
			con->packet_queue = g_list_remove(con->packet_queue, packet_to_send);
			free(packet_to_send);
		} else {
			con->total_dropped_to_send++;
#ifdef SOLARIS
			if (!((con->state == GS_WAITINGCONNECT && errno == ENOTCONN) || errno == EAGAIN)) {
#else
			if (errno != EAGAIN) {
#endif
				con->state = GS_FAILED;
				gui_update_con_status(con);
			}
			break;
		}
		
		iter = g_list_first(con->packet_queue);
	}

	return 1;
}

int con_send_to_all(char *buf, int len)
{
	GList *iter;
	struct newtella_connection *con;

        /* forward a packet every alive con */
	iter = g_list_first(gl_con);

	while (iter) {
		con = iter->data;
		
		if (con->state == GS_CONNECTED)
			con_send_packet(con, buf, len);
		
		iter = g_list_next(iter);
	}

	return 1;
}

int con_add(guint32 ip, guint16 port)
{
	struct newtella_connection *new_con = newtella_malloc(sizeof(struct newtella_connection));
	
	new_con->ip = ip;
	new_con->port = port;
	new_con->total_sent = 0;
	new_con->total_dropped_to_send = 0;
	new_con->total_handled = 0;
	new_con->total_received_dropped = 0;
	new_con->total_bad_received = 0;
	new_con->packet_queue = NULL;
	new_con->queue_size = 0;
	new_con->buffer_len = 0;
	new_con->state = GS_WAITINGCONNECT;	
	new_con->id = ++gl_options->current_con;

	if (gl_options->current_con == 1)
		gl_con = g_list_append(NULL, new_con);
	else
		gl_con = g_list_append(gl_con, new_con);

	gui_add_con_to_list(new_con); 

	new_con->s = newtella_connect(ip, port);
	
	new_con->read_tag = gdk_input_add(new_con->s, GDK_INPUT_READ | GDK_INPUT_EXCEPTION, 
		      (GdkInputFunction) on_gnutellanet_read, (gpointer)new_con);
	
	new_con->write_tag = gdk_input_add(new_con->s, GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION, 
		      (GdkInputFunction) on_gnutellanet_write, (gpointer)new_con);
	
	
	if (new_con->s == -1) {
		new_con->state = GS_FAILED;
		gui_update_con_status(new_con);
		return (-1);
	}
	
	con_send_packet(new_con, "GNUTELLA CONNECT/0.4\n\n", strlen("GNUTELLA CONNECT/0.4\n\n"));
	
	/* update time */
	new_con->last_time = time((time_t)NULL);
	
	return(1);
}

int con_dec(void)
{
	struct newtella_connection *con;
	GList *p;
	
	p = g_list_last(gl_con);
	con = p->data;

	con->state = GS_FAILED;
	
	return 1;
}

int con_remove(struct newtella_connection *con)
{
	/* remove host */
	host_remove(con->ip, con->port);

	gl_con = g_list_first(gl_con);
	
	gl_con = g_list_remove(gl_con, con);
	
	if (!g_list_length(gl_con))
		gl_con = NULL;

	gui_remove_con_from_list(con->id-1);
	gl_options->current_con--;

	gdk_input_remove(con->read_tag);
	gdk_input_remove(con->write_tag);
	
	close(con->s);
	newtella_free(con);

	con_fix_ids();

	gui_dump_con(); 
	return(1);
}

int con_down_remove(struct newtella_connection *d_con)
{
	gl_down_con = g_list_remove(gl_down_con, d_con);
       
	gui_remove_down_con_from_list(d_con->id-1);
	gl_options->current_down_con--;

	gdk_input_remove(d_con->read_tag);
	
	close(d_con->s);
	newtella_free(d_con);
	
	con_down_fix_ids(); 

	return 1;
}

int con_up_remove(struct newtella_connection *u_con)
{
	gui_update_up_con(u_con);

	gl_up_con = g_list_remove(gl_up_con, u_con);
       
	gl_options->current_up_con--;

	gdk_input_remove(u_con->write_tag);
	
	close(u_con->s);
	newtella_free(u_con);
	
	return 1;
}

int con_reset_buffer(struct newtella_connection *con)
{
	con->buffer_len=0;
	con->b=NULL;
	
	return 1;
}

int con_activate(void)
{
	GList *host_iter, *con_iter;
	struct host_info *host_to_con; 
	struct newtella_connection *con; 
	int already_con = 0;
       
	gl_hosts_in_cache = g_list_first(gl_hosts_in_cache); 
	host_iter = gl_hosts_in_cache;
	while (host_iter != NULL) {
		host_to_con = gl_hosts_in_cache->data;
	
		already_con = 0;
		/* check if we are already connected to that host */
		gl_con = g_list_first(gl_con);
		con_iter = gl_con;
		while (con_iter != NULL) {
			con = gl_con->data;
			
			if (host_to_con->ip == con->ip &&
			    host_to_con->port == con->port) {
				already_con = 1;
				break;
			}
		
			con_iter = g_list_next(gl_con);
			if (con_iter != NULL)
				gl_con = con_iter;
		}
	
		if (!already_con) {	
			if (con_add(host_to_con->ip, host_to_con->port) != -1) 
				return 1;
			else
				/* we cannot connect to host, so we 
				   remove it from cache. */
				gl_hosts_in_cache = g_list_remove(gl_hosts_in_cache, host_to_con);
		}
	       
		host_iter = g_list_next(gl_hosts_in_cache);
		if (host_iter != NULL)
			gl_hosts_in_cache = host_iter;
	}
	
	return 1;
}

int con_fix_ids(void)
{
	GList *iter;
	int i = 1;
	struct newtella_connection *con;

	iter = g_list_first(gl_con);
       
	while (iter != NULL) {
		con = iter->data;
			
		con->id = i++;	
		
		iter = g_list_next(iter);
		
	}

	return 1;
}

int con_down_fix_ids(void)
{
	GList *iter;
	int i = 1;
	struct newtella_connection *con;

	iter = g_list_first(gl_down_con);
       
	while (iter != NULL) {
		con = iter->data;
			
		con->id = i++;	
		
		iter = g_list_next(iter);
		
	}

	return 1;
}
