/*-*-linux-c-*-*/

/*
 *  gnewtellium - Newtella for Unix
 *  Copyright (C) 2001 Elias Athanasopoulos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>

#include "global.h"
#include "newtella.h"
#include "gnutella_header.h"
#include "gui.h"
#include "config.h"
#include "file.h"
#include "uploads.h"
#include "downloads.h"
#include "hosts.h"

extern struct newtella_options *gl_options;
extern int gl_server, gl_read_tag;
extern struct host_info *gl_my_host_info;

gchar *ip2str(guint32 ip)
{
	/* taken from gtk-gnutella */
        struct in_addr ia;
        ia.s_addr = g_htonl(ip);
        return inet_ntoa(ia);
}


guint32 str2ip(gchar *str)
{
        /* taken from gtk-gnutella */
        struct in_addr ia;
        gint r;
        r = inet_aton(str, &ia);
        if (r) return g_ntohl(ia.s_addr);
        return 0;
}

int is_mp3(gchar *str)
{
	return (!g_strncasecmp(str+strlen(str)-4, ".mp3", 4));
}

int newtella_server(int port)
{
	int rval, s;
	struct sockaddr_in server;

	s = socket(AF_INET, SOCK_STREAM, 0);
	
	if (s < 0) {
		perror("server socket");
		return 0;
	}
	
	/* non blocked */
	fcntl(s, F_SETFL, O_NONBLOCK);

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = htonl(INADDR_ANY);
	server.sin_port = htons(port);

	rval = bind(s, (struct sockaddr_in *)&server, sizeof(server));

	if (rval < 0) {
		perror("server bind");
		return 0;
	}
	
	return s;
}

int newtella_init_server(guint16 port)
{
	gl_server = newtella_server(port);
	
	if (!gl_server) 
		return 0;

	if (listen(gl_server, 10) < 0) 
                perror("server listen");
        

	gl_read_tag = gdk_input_add(gl_server, GDK_INPUT_READ | GDK_INPUT_EXCEPTION, 
				    (GdkInputFunction) newtella_welcome_client, NULL);
        
	return 1;
}

int newtella_kill_server(void)
{
	/* try to kill server */
	if (gl_server) {
		gdk_input_remove(gl_read_tag);
		close(gl_server);
	}

	return 1;
}

int newtella_connect(guint32 ip, guint16 port)
{
	struct sockaddr_in server;
	int s;

	s = socket(AF_INET, SOCK_STREAM, 0); 
	
	fcntl(s, F_SETFL, O_NONBLOCK);
	
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = ip;

	if (connect(s, &server, sizeof(server)) < 0) {
		if (errno != EINPROGRESS) {
			perror("connect (client)");
			return (-1);
		} 
	}

	return s;
}

int newtella_welcome_client(gpointer gdata, gint s, GdkInputCondition cond)
{
	struct sockaddr_in addr;
	int sock, len = sizeof(struct sockaddr_in);
	struct newtella_connection *in_con;

	if (cond == GDK_INPUT_EXCEPTION) 
		return -1;
 
	sock = accept(s, (struct sockaddr *) &addr, &len);

	if (sock == -1)
		return -1;

	/* make it non-block */
	fcntl(sock, F_SETFL, O_NONBLOCK);

	/* create a new incoming connection */
	in_con = newtella_malloc(sizeof(struct newtella_connection));
	in_con->s = sock;
	in_con->ip = g_ntohl(addr.sin_addr.s_addr);
        in_con->port = g_ntohs(addr.sin_port);
        in_con->total_sent = 0;
        in_con->total_dropped_to_send = 0;
        in_con->total_handled = 0;
        in_con->total_received_dropped = 0;
        in_con->total_bad_received = 0;
        in_con->packet_queue = NULL;
        in_con->queue_size = 0;
        in_con->buffer_len = 0;
        in_con->state = GS_WAITINGCONNECT;     
        in_con->id = ++gl_options->current_con;

        if (gl_options->current_con == 1)
                gl_con = g_list_append(NULL, in_con);
        else
                gl_con = g_list_append(gl_con, in_con);

	gui_add_con_to_list(in_con); 
        
        in_con->read_tag = gdk_input_add(in_con->s, GDK_INPUT_READ | GDK_INPUT_EXCEPTION, 
					 (GdkInputFunction) on_gnutellanet_read, (gpointer)in_con);
        
        in_con->write_tag = gdk_input_add(in_con->s, GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION, 
					  (GdkInputFunction) on_gnutellanet_write, (gpointer)in_con);
	/* update time */
        in_con->last_time = time((time_t)NULL);

	return 1;
}

int newtella_packet_sucks(char *buf, int len)
{
	struct gnutella_header *gh;
	
	/* handshake/download/upload data */
	if (!memcmp(buf, "GNUTELLA OK\n\n", strlen("GNUTELLA OK\n\n")))
		return 0;
	if (!memcmp(buf, "GNUTELLA CONNECT/0.4\n\n", strlen("GNUTELLA CONNECT/0.4\n\n")))
		return 0; 
	if (!memcmp(buf, "GET /get/", strlen("GET /get/")))
		return 0;
	if (!memcmp(buf, "GIV", strlen("GIV")))
		return 0;

	/* we have a gnutella package, let's see
	   if it is a valid one. 
	*/
	gh = newtella_malloc(GN_HEADER_SIZE);
	serialize_header(gh, buf);	

	if ((gh->length > MAX_PACKET_LENGTH) || (gh->length < 0))
	{
		goto bad;
	}
	
	/* length check */
	switch (gh->function) {
	case GNUTELLA_INIT:
		if (gh->length != 0) goto bad; 
		break;
	case GNUTELLA_INITRESPONSE:;
		if (gh->length != GN_PONG_SIZE) goto bad; 
		break;
	case GNUTELLA_PUSHREQUEST:
		if (gh->length != GN_PUSH_SIZE) goto bad;
		break;
	case GNUTELLA_SEARCH:
		if (gh->length < (GN_QUERY_SIZE+2)) goto bad;
		break;
	case GNUTELLA_SEARCHRESPONSE:
		break;
	default:
		/* ok, this is defenately a bad package */
		goto bad;
	}
	
	/*
	if (gh->function == GNUTELLA_SEARCH) {
		if (gh->length < GN_QUERY_SIZE+2) 
			goto bad;

		if (gh->length > GNUTELLA_MAX_QUERY) 
			goto bad;

		m = buf + GN_HEADER_SIZE + gh->length-1;
		if (*m != '\0') 
			goto bad;

		m = buf + GN_HEADER_SIZE + GN_QUERY_SIZE;

		if (!strlen(m)) 
			goto bad;
	}
	*/

	newtella_free(gh);
	return 0;

 bad:

	newtella_free(gh);	
	return 1;
}

int handle_connect(struct newtella_connection *con, char *buf)
{

	if (memcmp(buf, "GNUTELLA CONNECT/0.4\n\n", strlen("GNUTELLA CONNECT/0.4\n\n"))) 
		return 0;

	con->last_time = time((time_t)NULL);
	con->state = GS_CONNECTED;
	
	con_send_packet(con, "GNUTELLA OK\n\n", strlen("GNUTELLA OK\n\n")); 
       
	return 1;
}

int handle_connect_response(struct newtella_connection *con, char *buf)
{
	if (memcmp(buf, "GNUTELLA OK\n\n", strlen("GNUTELLA OK\n\n")))
		return 0;

	con->last_time = time((time_t)NULL);
	con->state = GS_CONNECTED;

	gnutella_send_ping(con);

	return 1;
}

int handle_get_request(struct newtella_connection *con, char *buf)
{
	if (memcmp(buf, "GET /get/", 8))
		return 0;

	con->last_time = time((time_t)NULL);
	con->state = GS_CONNECTED;

	return 1;
}

int handle_push(struct newtella_connection *con, char *buf)
{
	struct gnutella_push *gpush = newtella_malloc(sizeof(struct gnutella_push));

	serialize_push(gpush, buf);

	/* is it for us? */
	if (!memcmp(&(gl_options->gl_guid), &(gpush->_guid), 16)) {
		upload_add(gpush->ip, gpush->port, gpush->index, &(gpush->_guid));
	} /* else we must route the push */
	   
	con->total_handled++;
     	
 	return GN_PUSH_SIZE;
}

int handle_init(struct newtella_connection *con, struct gnutella_header *gh)
{
	GList *iter;
	struct routing_info *rp;

	if (gh->TTL < 1 || gh->hops > GNUTELLA_MAX_HOPS) {
		con->total_bad_received++;
		return 0;
	} else 
		con->total_handled++;
	
	/* check if we have seen the ping */
	iter = g_list_first(gl_route_pings);
	while (iter != NULL) {
		rp = (struct routing_info *) iter->data;
		
		if (!memcmp(&rp->_guid, &gh->_guid, sizeof(guid)))
			return 0;

		iter = g_list_next(iter);
	}
 
	/* send the reply (pong) */
	gnutella_send_pong(con, gh->_guid);
	
        /* store the new ping */
	rp = newtella_malloc(sizeof(struct routing_info));
	rp->f = GNUTELLA_INIT;
	rp->_guid = gh->_guid;
	rp->time = time((time_t *)NULL);
	rp->con = con;
	
	gl_route_pings = g_list_append(gl_route_pings, (struct routing_info *)rp);
	 
	/* route the ping */
	gnutella_fwd_ping(gh, con);

	return 0;
}

int handle_init_response(struct newtella_connection *con, char *buf, struct gnutella_header *gh)
{
	struct gnutella_pong_header *gpong;
	struct routing_info *rp;
	GList *iter;
	
	if (gh->TTL < 1 || gh->hops > GNUTELLA_MAX_HOPS) {
		con->total_bad_received++;
		return GN_PONG_SIZE;
	} else {
		con->total_handled++;
		gpong = newtella_malloc(GN_PONG_SIZE);
		serialize_pong(gpong, buf);
		host_add_in_cache(gpong);
	}
	
	/* check if we have to route the pong */
	iter = g_list_first(gl_route_pings);
	while (iter != NULL) {
		rp = (struct routing_info *) iter->data;

		if (!memcmp(&rp->_guid, &gh->_guid, sizeof(guid))) {
			/* pass the pong */
			gnutella_fwd_pong(rp->con, gh, gpong);
			break;
		} 
		
		iter = g_list_next(iter);
	}
 
	newtella_free(gpong);
	
	return GN_PONG_SIZE;
}

int handle_search(struct newtella_connection *con, struct gnutella_header *gh, char *buf)
{
	struct gnutella_query_header *gquery = newtella_malloc(GN_QUERY_SIZE);
	GList *keywords, *iter;
	GList *found_hits;

	con->total_handled++;

	memcpy(&(gquery->min_speed), buf, 2);

	keywords = file_extract_keywords(buf+2);
	
	if (keywords) {
		iter = g_list_first(keywords);
		while (iter) {
			found_hits = file_scan_db((char *)iter->data);
			
			/* ok we have what he/she wants */
			if (found_hits) {
				gnutella_send_query_response(con, found_hits, gh->_guid);
				g_list_free(found_hits);
			}
			
			iter = iter->next;
		}
	}

	/* route the query */
	gquery->name = newtella_malloc(gh->length-1);
	strcpy(gquery->name, buf+2); 
       
	gnutella_fwd_query(gh, gquery, con); 
	 
	newtella_free(gquery->name); 
 	newtella_free(gquery);
	return gh->length;
}

int handle_search_response(struct newtella_connection *con, struct gnutella_header *gh, char *buf)
{
	struct gnutella_hits *ghits = newtella_malloc(GN_HITS_SIZE);
	struct gnutella_hit *ghit;
	int i, last_found = gl_options->files_found;
	struct file_found *found_hit;
	gchar *dummy[6] = { "", "", "", "", "", "" };
	guid guid_for_push;

	con->total_handled++;

	/* store the guid to use it for push */
	memcpy(&guid_for_push, buf+gh->length-16, 16);

	serialize_hits(ghits, buf);
				
	ghit = newtella_malloc(GN_HIT_SIZE);
	buf += GN_HITS_SIZE;

	for (i=0; i<ghits->hits; i++) {

		/* check if we reached the maximum results limit */
		if (gl_options->max_search_results == gl_options->files_found)
			goto skip;

		serialize_hit(ghit, buf);

		/* check if we really have mp3 results */
		if (file_is_mp3(ghit->name)) {
			found_hit = newtella_malloc(sizeof(struct file_found));
			found_hit->file_name = newtella_malloc(strlen(ghit->name)+1);
			strcpy(found_hit->file_name, ghit->name);
			newtella_free(ghit->name);

			found_hit->_guid = guid_for_push;
			found_hit->index = ghit->index;
			found_hit->size = ghit->size;
			found_hit->port = ghits->port;
			found_hit->ip = ghits->ip;
			found_hit->speed = ghits->speed;
			found_hit->received = 0;
			
			gl_results = g_list_append(gl_results, found_hit);
			gl_options->files_found++;

			gtk_clist_append(GTK_CLIST(search_list), dummy);
			
			/* go to next hit */
			buf += GN_HIT_SIZE + strlen(found_hit->file_name) + 2;
		} else { 
			buf += GN_HIT_SIZE + strlen(ghit->name) + 2;
			newtella_free (ghit->name);
		}
	}

 skip: 	
	newtella_free(ghit);
	newtella_free(ghits);
	
	gui_update_found_results(last_found);
 
	return gh->length;
}

int handle_packet(struct newtella_connection *con, struct gnutella_header *gh, char *buf)
{
	switch (gh->function) {
	case GNUTELLA_INIT:
		return handle_init(con, gh);
		break;
	case GNUTELLA_INITRESPONSE:
		return handle_init_response(con, buf, gh);
		break;
	case GNUTELLA_PUSHREQUEST:
		return handle_push(con, buf);
		break;
	case GNUTELLA_SEARCH:
		return handle_search(con, gh, buf);
		break;
	case GNUTELLA_SEARCHRESPONSE:
		return handle_search_response(con, gh, buf);
		break;
	default:
		break;
	}
	
	return 0;
}

int on_gnutellanet_write(gpointer gdata, gint s, GdkInputCondition cond)
{
	struct newtella_connection *con = (struct newtella_connection *)gdata;

	if (cond == GDK_INPUT_EXCEPTION) {
		con->state = GS_FAILED;
		return -1;
	}
	
	if (con->state == GS_CONNECTED || con_send_from_queue(con)) {
		gdk_input_remove (con->write_tag); 
		return 1;
	}

	return 1;
}

int on_gnutellanet_read(gpointer gdata, gint s, GdkInputCondition cond)
{
	struct newtella_connection *con = (struct newtella_connection *)gdata;
	struct gnutella_header *gh;
	guint16 bytes_readed = 0, bytes_to_read = 0, 
		bytes_parsed = 0, bytes_left = 0,
		bytes_done = 0;
	char *p;

	con->last_time = time((time_t)NULL);

	if (cond == GDK_INPUT_EXCEPTION) {
		con->state = GS_FAILED;
		return -1;
	}
			
	bytes_to_read = bytes_avail(s);
	if (!bytes_to_read) {
		con->state = GS_FAILED;
		return -1;
	}

	/*con->buffer = g_realloc(con->buffer, con->buffer_len + bytes_to_read);*/
	con->b = con->buffer + con->buffer_len;
	con->buffer_len += bytes_to_read;
	bytes_readed = con_recv_packet(con, bytes_to_read);
	
	p = con->buffer;
	gh = newtella_malloc(GN_HEADER_SIZE);
      
	do {
		bytes_left = con->buffer_len - bytes_parsed;

		if (newtella_packet_sucks(p, bytes_left))
		{
			con->state = GS_FAILED;
		}
		
	
		switch (con->state) {
		case GS_FAILED:
			newtella_free(gh);
			return 1;
		case GS_CONNECTED:						
			if (bytes_left >= GN_HEADER_SIZE)
				serialize_header(gh, p);
			else {	
				memcpy(con->buffer, p, bytes_left);
				con->buffer_len = bytes_left;
				goto skip;
			}
			/*
			if (gh->length > (bytes_left-GN_HEADER_SIZE)) {
				printf("%d %d\n", gh->length, bytes_left);
				memcpy(con->buffer, p, bytes_left);
				con->buffer_len = bytes_left;
				goto skip;
			}
			*/
			
			p += GN_HEADER_SIZE;
			bytes_parsed += GN_HEADER_SIZE;
			
			if (bytes_left > gl_options->con_speed*BYTES_PER_SECOND*3 &&
			    (gh->function == GNUTELLA_INIT || gh->function == GNUTELLA_SEARCH)) {
				con->total_received_dropped++;
				bytes_done += gh->length;
			} else if (bytes_left > MAX_BYTES_PER_SECOND*7) {
				con->total_received_dropped++;
				bytes_done += gh->length;
			} else
				bytes_done = handle_packet(con, gh, p);
			
			p += bytes_done;
			bytes_parsed += bytes_done;
			break;
		case GS_WAITINGCONNECT:
			/* incoming connection waiting to establish */
			if (handle_connect_response(con, con->buffer)) { 
				p += strlen("GNUTELLA OK\n\n");
				bytes_parsed += strlen("GNUTELLA OK\n\n");
				break;
			} /* outgoing connection waiting to establish */
			else if (handle_connect(con, con->buffer)) {
				p += strlen("GNUTELLA CONNECT/0.4\n\n");
				bytes_parsed += strlen("GNUTELLA CONNECT/0.4\n\n");
				break;
			} /* download connection waiting to establish */
			else if (handle_get_request(con, con->buffer)) {
				upload_start(con, p);
				break;
			} /* crap */ 
			else if (con->buffer_len > strlen("GNUTELLA CONNECT/0.4\n\n")) {
				con->state = GS_FAILED;
				break;
			} 
			bytes_parsed = con->buffer_len;
			break;
		default:
			break;
		}
		
	} while (bytes_parsed < con->buffer_len);

	if (bytes_parsed == con->buffer_len)
		con_reset_buffer(con);

 skip:	
	
	newtella_free(gh);
	return 1;
}

int newtella_sched(void)
{
	GList *iter;
	struct newtella_connection *con;
	int i;

	if (gl_options->current_con) 
		gui_dump_con();
	if (gl_options->current_con < gl_options->min_con && gl_options->auto_connect)
		for (i=0; i<(gl_options->min_con - gl_options->current_con); i++)
			con_activate();
	if (gl_options->current_con > gl_options->max_con && gl_options->auto_connect)
		for (i=0; i<(gl_options->current_con - gl_options->max_con); i++)
			con_dec();
	
	/* download connections */
	iter = g_list_first(gl_down_con);
	while (iter != NULL) {
		con = iter->data;
		
		/*if ((time((time_t)NULL) - con->last_time) > 120) {
			con_remove(con);
			break;
			}*/
		
		if (con->state == GS_FAILED) {
			con_down_remove(con);
			break;
		}
		
		iter = g_list_next(iter);
	}
	
	/* ordinary connections */
	iter = g_list_first(gl_con);
	while (iter != NULL) {
		con = iter->data;
		if (con->state == GS_FAILED || con->state == GS_TIMEOUT) {
			con_remove(con);
			break;
		}

		if ((time((time_t)NULL) - con->last_time) > 30) {
			con->state = GS_TIMEOUT;
			gui_update_con_status(con);
			break;
		}
		
		if (con->state == GS_CONNECTED)
			con_send_from_queue(con);
	       

		iter = g_list_next(iter);
	}

	return 1;
}


